#include <iostream>
#include <fstream>
#include <vector>
#include <regex>
#include <unistd.h>
#include "errors.h"
using namespace std;

istream& operator >>( istream&, string& );            // ввод всей программы в строку 
regex pattern( string& );                             // трансляция регулярного выражения
ostream& operator <<( ostream&, smatch& );            // диагностика отождествления
void parse( const string& );                          // разбивка строки на лексемы

int debug_level = 0;                                  // уровень отладочного вывода 
static vector<string> lex;                            // итоговый результат разбора

int main( int argc, char *argv[] ) {
   char sopt[] = "sv", c;
   bool strong = false;
   while( -1 != ( c = getopt( argc, argv, sopt ) ) )
      switch( c ) {
         case 's':
            strong = true;
            break;
         case 'v':
            debug_level++;
            break;
         default :
            cerr << "формат команды: допустимы только опции: -" << sopt << endl;
            return 1;
      }
   if( optind == argc ) {
       cerr << "формат: " << argv[ 0 ] << " <файлы кода...>" << endl;
       return 2;
   }
   string sp = R"(^program\s*(\{.*)\s*$)";            // "^(program)\\s*(\\{.*)\\s*$"
   regex r = pattern( sp );
   ifstream fin;
   smatch m; 
   for( int i = optind; i < argc; i++ ) {
      fin.open( argv[ i ] );
      if( !fin ) {
         cerr << "файл программы не найден: " << argv[ i ] << endl;
         if( strong ) return 3; else continue;
      }
      string text;
      fin >> text;
      fin.close();
      if( debug_level > 2 ) cerr << text << endl;
      if( regex_match( text, m, r ) ) {
         if( debug_level > 2 ) cerr << m;
         lex.clear();
         try { parse( m[ 1 ].str() ); }
         catch( syn_exception& e ) {                  // ошибки выполнения
            cerr << (const char*)e << endl;
            if( strong ) return 4; else continue;
         }         
         if( debug_level > 0 ) cerr << argv[ i ] << " - итог трансляции:" << endl;
         for( string &s: lex ) cout << s << endl;     // только результат cout << ...
      }
      else {
         cerr << "ошибка формата программы " << argv[ i ] << endl;
         if( strong ) return 5; else continue;
      }
   }
}

istream& operator >>( istream& inp, string& text ) { 
   text = "";
   while( true ) {  
      string cur;
      getline( inp, cur );
      if( !inp ) break;                               // EOF
      if( debug_level > 0 ) cerr << cur << endl;
      int cmnt;
      if( ( cmnt = cur.find( "//" ) ) >= 0 )          // комментарии 
         cur = cur.substr( 0, cmnt );
      if( !cur.length() ) continue;                   // пустые строки
      text += cur;
   }
   return inp;
}

// static constexpr flag_type basic = regex_constants::basic;
// static constexpr flag_type extended = regex_constants::extended;
// static constexpr flag_type awk = regex_constants::awk;
// static constexpr flag_type grep = regex_constants::grep;
inline regex pattern( string& r ) {
   try { return regex( r ); } 
   catch( regex_error& e ) {                          // ошибка записи регулярного выражения
      cerr << "синтаксис регулярного выражения:\t" << r << endl;
      exit( 3 );
   }
}

inline ostream& operator <<( ostream& out, smatch& m ) {
   out << "match [" << m.size() << "]:" << endl;
   for( unsigned i = 1; i < m.size(); i++ )
      cerr << "[" << m.position( i ) << ":" << m.length( i ) << "]\t[" << m[ i ].str() << "]" << endl; 
   return out;
}

string trim( const string& s ) {   
   string sp = R"(^\s*(.*?)\s+$)", t = s + " ";
   regex r = pattern( sp );
   smatch m;
   regex_match( t, m, r );
   return m[ 1 ].str(); 
}

//=================================================================================================

string next_label( void ) {
   static int loc_label = 1;                          // тек. локальная метка
   const int len = 4;
   char s[ len + 1 ];
   sprintf( s, "%04d", loc_label++ );
   return string( s ); 
}

int block( const string& s, size_t b ) {              // нахождение границы вложенных блоков
   const string delim[ 2 ] = { "([{", ")]}" };
   int n;
   if( ( n = delim[ 0 ].find( s[ b ] ) ) < 0 )
      return b;
   char cbeg = delim[ 0 ][ n ], cend = delim[ 1 ][ n ];
   n = 1;
   while( ++b < s.length() ) {
      if( s[ b ] == cbeg ) n++;
      if( s[ b ] == cend ) n--;
      if( 0 == n ) return b;      
   }
   return -1;
}

//-------------------------------------------------------------------------------------------------

// string s1 = "asdf", s2 = "fdsa"; s1 = ""; => <string>[0:6] | <s1 = "asdf", s2 = "fdsa">[7:24] | < s1 = "";>[32:9]
string eval_var( smatch& m ) {
   lex.push_back( m[ 1 ].str() + " " + m[ 2 ].str() );
   return m[ 3 ].str(); 
}

// { { a = b + 1; } b = c - 1; } => <{ { a = b + 1; } b = c - 1; }>[0:29]
string eval_block( smatch& m ) {
   int n = block( m[ 1 ].str(), 0 );
   string oper;
   if( n > 0 ) {
      oper = trim( m[ 1 ].str().substr( 1, n - 1 ) ); // внутренность блока
      parse( oper );                                  // рекурсивная обработка блока 
   }
   return n + 1 < (int)m[ 1 ].str().length() ? m[ 1 ].str().substr( n + 1 ) : ""; 
}


// ^[ ]*if[ ]*[\(]{1}(.*)(\){1}[ ]*\{)(.*?)\}[ ]*else[ ]*\{(.*)$
//        if ( i > j )   { x = i ; } else { x = j ; }   ; y ; z ;
//            1......2....3........        4.....................
//1: [ i > j ] <11:7>
//2: [)   {] <18:5>
//3: [ x = i ; ] <23:9>
//4: [ x = j ; }   ; y ; z ;] <40:22>
string eval_else( smatch& m ) {
   string labl = next_label(), blbl = "B" + labl, flbl = "F" + labl;
   lex.push_back( trim( m[ 1 ].str() ) + " ifn " + blbl );
   parse( trim( m[ 3 ].str() ) );            // блок then
   lex.push_back( "goto " + flbl );
   lex.push_back( blbl + ":" );
   int b = m.position( 4 ) - 1,   
       n = block( m[ 0 ].str(), b );
   string tmp;
   if( n >= b ) {
      tmp = m[ 0 ].str().substr( b + 1, n - b - 1 );
      parse( tmp );                          // блок else
   }
   else
      throw syn_exception( "if: баланс скобок else {...}" );
   lex.push_back( flbl + ":" );
   return n + 1 < (int)m[ 0 ].str().length() ? m[ 0 ].str().substr( n + 1 ) : ""; 
}

// ^[ ]*if[ ]*[\(]{1}(.*)(\){1}[ ]*\{)(.*)$
//         if ( i > j )   { x = i ; } ; x = 3
//            1......2....3.................
//1: [ i > j ] <11:7>
//2: [)   {] <18:5>
//3: [ x = i ; } ; x = 3   ] <23:21>
string eval_if( smatch& m ) {
   string labl = next_label(), flbl = "F" + labl;
   lex.push_back( trim( m[ 1 ].str() ) + " ifn " + flbl );
   int b = m.position( 3 ) - 1,   
       n = block( m[ 0 ].str(), b );         // блок операторов
   string tmp;
   if( n >= b ) {
      tmp = m[ 0 ].str().substr( b + 1, n - b - 1 );
      parse( tmp );
   }
   else
      throw syn_exception( "if: баланс скобок {...}" );
   lex.push_back( flbl + ":" );
   return n + 1 < (int)m[ 0 ].str().length() ? m[ 0 ].str().substr( n + 1 ) : ""; 
}

// ^[ ]*for[ ]*[\(]{1}([^\)|^;]*);{1}([^\)|^;]*);{1}[^\)]*(\){1}[ ]*\{)(.*)$
//                    1......        2..........          3........... 4....
//     for ( i = 0; i < 5; i = i + 1 ) { i = i + 1; } ; x = 0;
//          1..... 2.....            3..4.....................
//1: [ i = 0] <9:6>
//2: [ i < 5] <16:6>
//3: [) {] <34:3>
//4: [ i = i + 1; } ; x = 0; ] <37:23>
string eval_for( smatch& m ) {
   lex.push_back( trim( m[ 1 ].str() ) );    // начальное значение
   string labl = next_label(), blbl = "B" + labl, flbl = "F" + labl;
   lex.push_back( blbl + ": " + trim( m[ 2 ].str() ) + " ifn " + flbl );
   int b = m.position( 4 ) - 1,   
       n = block( m[ 0 ].str(), b );         // блок операторов 
   string tmp;
   if( n >= b ) {
      tmp = m[ 0 ].str().substr( b + 1, n - b - 1 );
      parse( tmp );
   }
   else
      throw syn_exception( "for: баланс скобок {...}" );
   int ps = m.position( 2 ) + m.length( 2 ) + 2,
       ls = m.position( 3 ) - ps;
   lex.push_back( m[ 0 ].str().substr( ps, ls ) );
   lex.push_back( "goto " + blbl );
   lex.push_back( flbl + ":" );
   return n + 1 < (int)m[ 0 ].str().length() ? m[ 0 ].str().substr( n + 1 ) : ""; 
}

// while ( a < 3 ) { a = a + 1 }; a = 5 ; => <while (>[0:7] | < a < 3 ) { a = a + 1 }; a = 5 ;>[7:31]
string eval_while( smatch& m ) {
   string labl = next_label(), blbl = "B" + labl, flbl = "F" + labl;
   int b = m.position( 1 ) + m.length( 1 ) - 1,
       n = block( m[ 0 ].str(), b );
   string cond;
   if( n > b )
      cond = trim( m[ 0 ].str().substr( b + 1, n - b - 1 ) );
   else
      throw syn_exception( "while: баланс скобок (...)" );
   lex.push_back( blbl + ": " + cond + " ifn " + flbl );
   string last = trim( m[ 0 ].str().substr( n + 1 ) );
   n = block( last, 0 );
   if( n > 0 ) 
      parse( trim( last.substr( 1, n - 1 ) ) );       // рекурсивная обработка операторов блока 
   else
      throw syn_exception( "while: баланс скобок {...}" );
   lex.push_back( "goto " + blbl );
   lex.push_back( flbl + ":" );
   return n + 1 < (int)last.length() ? last.substr( n + 1 ) : ""; 
}

// do { a = a + 1 } while( a< 5 )  ; a = 5; => <do >[0:3] | <{ a = a + 1 } >[3:14] | <( a< 5 )  ; a = 5;>[22:18]
string eval_do( smatch& m ) { 
   string blbl = "B" + next_label();
   lex.push_back( blbl + ": " );
   int n = block( m[ 2 ].str(), 0 );
   string oper, cond;
   if( n > 0 )
      parse( oper = trim( m[ 2 ].str().substr( 1, n - 1 ) ) );
   else 
      throw syn_exception( "do: баланс скобок {...}" );
   n = block( m[ 3 ].str(), 0 );
   if( n > 0 )
      lex.push_back( trim( m[ 3 ].str().substr( 1, n - 1 ) ) + " if " + blbl );
   else 
      throw syn_exception( "do: баланс скобок (...)" );
   return n + 1 < (int)m[ 3 ].str().length() ? m[ 3 ].str().substr( n + 1 ) : ""; 
}

// read( a ); a = a + 1; => <read( a )>[0:9] | < a = a + 1;>[10:11]
string eval_single( smatch& m ) {
   lex.push_back( m[ 1 ].str() );
   return m[ 2 ].str(); 
}

string eval_norel( smatch& m ) {
   cerr << "не реализовано или ошибка: " << m[ 1 ].str().substr( 0, 10 ) << "..." << endl;
   return m[ 1 ].str().substr( 1 ); 
}

void parse( const string& text ) {
   static int level = 0;                              // уровень рекурсии
   level++;
   string src( text );
   if( debug_level > 1 ) cerr << "begin parsing [" << level << "]:" << endl;
   string shift;
   for( int i = 0; i < level; i++ ) shift += "------";
   shift += " ";
   struct {                                          // структура описания лексемы
      string s;                                      // регулярное выражение выделения лексемы 
      string (*f)( smatch& );                        // функция трансляции лексемы
      string d;                                      // наименование (для отладки)
   } static pats[] = { 
/*описание*/     { R"(^(int|string)\s+([^;]*);{1}(.*)$)", eval_var, "описание" },
/*блок*/         { R"(^\s*(\{.*)$)", eval_block, "блок" },
/*пусто*/        { R"(^\s*;(.*)$)", []( smatch& m )->string { return m[ 1 ].str(); }, "пусто" },
/*if-else*/      { R"(^[ ]*if[ ]*[\(]{1}(.*)(\){1}[ ]*\{)(.*?)\}[ ]*else[ ]*\{(.*)$)", eval_else, "if-else" },
/*if*/           { R"(^[ ]*if[ ]*[\(]{1}(.*)(\){1}[ ]*\{)(.*)$)", eval_if, "if" },
//                 R"(^[ ]*if[ ]*[\(]{1}(.*)(\){1}[ ]*\{)(.*?)\}[ ]*else[ ]*\{(.*)$)"
/*for*/          { R"(^\s*for\s*[\(]{1}([^\)|^;]*);{1}([^\)|^;]*);{1}[^\)]*(\){1}\s*\{)(.*)$)", eval_for, "for" },
/*while*/        { R"(^\s*(while\s*\()(.*)$)", eval_while, "while" },
/*do*/           { R"(^\s*(do\s*)(\{.*)while\s*(\(.*)$)", eval_do, "do" },
/*оператор*/     { R"(^\s*([^;]*);{1}(.*)$)", eval_single, "оператор" },
/*???*/          { R"(^(.*)$)", eval_norel, "не реализовано" }
              };
   static const int size = sizeof( pats ) / sizeof( pats[ 0 ] );
   while( src.length() ) {
      smatch m = smatch {};
      src = trim( src );
      int i;
      for( i = 0; i < size; i++ ) {
         regex r = pattern( pats[ i ].s );
         try {
            if( regex_match( src, m, r ) ) {
               if( debug_level > 1 ) cerr << pats[ i ].d << " - " << m;
                  src = pats[ i ].f( m );
               break;                                // соответствие лексемы найдено
            }
         }
         catch(...) { throw; }
      }
   }
   if( debug_level > 1 ) cerr << "final parsing [" << level << "]: " << src << endl;             
   level--;
}
